import math

from swimlane.core.resolver import AppResolver
from swimlane.core.resources.record_revision import RecordRevision


class RecordRevisionAdapter(AppResolver):
    """Handles retrieval of Swimlane Record Revision resources"""

    def __init__(self, app, record):
        super(RecordRevisionAdapter, self).__init__(app)
        self.record = record

    def get_all(self):
        """Get all revisions for a single record.

        Returns:
            RecordRevision[]: All record revisions for the given record ID.
        """
        raw_revisions = self._swimlane.request('get',
                                               'app/{0}/record/{1}/history'.format(self._app.id, self.record.id)).json()
        return [RecordRevision(self._app, raw) for raw in raw_revisions]

    def get(self, revision_number):
        """Gets a specific record revision.

        Keyword Args:
            revision_number (float): Record revision number

        Returns:
            RecordRevision: The RecordRevision for the given revision number.
            Raises: When revision is not an integer, a float NOT ending in ".0", or is less than 1
        """
        if isinstance(revision_number, (int, float)):
            if revision_number > 0 and revision_number % math.floor(revision_number) == 0:
                record_revision_raw = self._swimlane.request('get',
                                                             'app/{0}/record/{1}/history/{2}'.format(self._app.id,
                                                                                                     self.record.id,
                                                                                                     revision_number)).json()
                return RecordRevision(self._app, record_revision_raw)


        raise ValueError('The revision number must be a positive whole number greater than 0')


def validate_str(value, key):
  if not isinstance(value, str):
    raise ValueError('{} must be a string value.'.format(key))
  if value.strip() == '':
    raise ValueError('{} must not be an empty string value.'.format(key))

def validate_str_format(value, key):
  not_allowed = "@#$%^&*()"
  if any(ch in not_allowed for ch in value):
    raise ValueError('{} is not of the proper format.'.format(key))


import itertools


class Cursor(object):

    def __init__(self):
        self._elements = []

    def __len__(self):
        return len(list(self._evaluate()))

    def __iter__(self):
        for element in self._evaluate():
            yield element

    def __getitem__(self, item):
        return self._evaluate()[item]

    def _evaluate(self):
        """Hook to allow lazy evaluation or retrieval of cursor's elements

        Defaults to simply returning list of self._elements
        """
        return self._elements


class PaginatedCursor(Cursor):
    """Handle paginated lists, exposes hooks to simplify retrieval and parsing of paginated data"""

    default_limit = 0
    default_page_size = 10
    default_page_start = None
    default_page_end = None

    def __init__(self, limit=default_limit, page_size=default_page_size,
                 page_start=default_page_start, page_end=default_page_end):
        super(PaginatedCursor, self).__init__()

        self.__limit = limit
        self.page_size = page_size
        self.page_start = page_start
        self.page_end = page_end

        if self.__limit:
            self.page_size = min(self.page_size, self.__limit)

        if self.page_start and self.page_start <= 0:
            raise ValueError('page_start should be greater than 0')

        if self.page_end and self.page_end <= 0:
            raise ValueError('page_end should be greater than 0')

        if (self.page_start and self.page_end) and (self.page_start > self.page_end):
            raise ValueError('page_end cannot be less than page_start')

        if (self.page_start or self.page_end) and self.__limit != 0:
            raise ValueError(' page_start or page_end param is applicable only when limit is 0')

    def _evaluate(self):
        """Lazily retrieve and paginate report results and build Record instances from returned data"""
        if self._elements:
            for element in self._elements:
                yield element
        else:
            # Determine pagination range based on parameters
            if self.page_start and self.page_end:
                page_range = range(self.page_start-1, self.page_end)
            elif self.page_start:
                page_range = itertools.count(self.page_start-1)
            elif self.page_end:
                page_range = range(0, self.page_end)
            else:
                page_range = itertools.count()

            for page in page_range:
                raw_elements = self._retrieve_raw_elements(page)

                for raw_element in raw_elements:
                    element = self._parse_raw_element(raw_element)
                    self._elements.append(element)
                    yield element

                    if self.__limit and len(self._elements) >= self.__limit:
                        break

                # Break conditions for ending pagination
                if any([
                    len(raw_elements) < self.page_size,
                    self.__limit and len(self._elements) >= self.__limit,
                    self.page_size == 0
                ]):
                    break

    def _retrieve_raw_elements(self, page):
        """Send request and return response for single page of data"""
        raise NotImplementedError

    def _parse_raw_element(self, raw_element):
        """Hook to override parsing individual raw elements just before yielding"""
        return raw_element



import weakref


class SwimlaneResolver(object):
    """Provides automatic weakref resolution for Swimlane client to avoid circular references and 
    memory leaks """

    def __init__(self, swimlane):
        self.__ref_swimlane = weakref.ref(swimlane) if swimlane else swimlane

    @property
    def _swimlane(self):
        """Transparently resolve the swimlane weakref"""
        if not self.__ref_swimlane:
            referent = self.__ref_swimlane
        else:
            referent = self.__ref_swimlane()
        if referent is None:
            raise ReferenceError("The swimlane object has been garbage collected. The Swimlane "
                                 "object uses weak references to avoid memory leaks. The object "
                                 "gets garbage collected when there are no more references to it. "
                                 "See the weakref documentation for more information.")
        return referent


class AppResolver(SwimlaneResolver):
    """Provides automatic weakref resolution for Swimlane client and App instance"""

    def __init__(self, app):
        super(AppResolver, self).__init__(app._swimlane)

        self.__ref_app = weakref.ref(app) if app else app

    @property
    def _app(self):
        if not self.__ref_app:
            referent = self.__ref_app
        else:
            referent = self.__ref_app()
        if referent is None:
            raise ReferenceError(
                "The App object of the Swimlane Object has been garbage collected. Both the App "
                "object and the Swimlane object use weak references to avoid memory leaks. The "
                "instance gets garbage collected when there are no more strong references to it. "
                "See the weakref documentation for more information.")
        return referent



def validate_str_list(value, key):
    if not value or not isinstance(value, list):
        raise ValueError('{} must be a non-empty list value'.format(key))
    
    for i in value:
        if not isinstance(i, str) or i.strip() == '':
            raise ValueError('{} must contain non-empty string values'.format(key))